@@ -106,7 +106,13 @@ g_tls_connection_gnutls_set_handshake_priority (GTlsConnectionGnutls *gnutls)
   GTlsConnectionGnutlsPrivate *priv = g_tls_connection_gnutls_get_instance_private (gnutls);
   int ret;
 
-  g_assert (priority);
+  if (!priority)
+    {
+      /* initialize_gnutls_priority() previously failed and printed a warning,
+       * so no need for further warnings here.
+       */
+      return;
+    }
 
   ret = gnutls_priority_set (priv->session, priority);
   if (ret != GNUTLS_E_SUCCESS)
@@ -127,7 +133,7 @@ update_credentials_cb (GObject    *gobject,
   int ret;
 
   database = g_tls_connection_get_database (G_TLS_CONNECTION (gnutls));
-  if (database)
+  if (database && G_IS_TLS_DATABASE_GNUTLS (database))
     {
       credentials = g_tls_database_gnutls_get_credentials (G_TLS_DATABASE_GNUTLS (database), &error);
       if (!credentials)
@@ -190,7 +196,7 @@ g_tls_connection_gnutls_initable_init (GInitable     *initable,
     flags |= GNUTLS_DATAGRAM;
 
   database = g_tls_connection_get_database (G_TLS_CONNECTION (gnutls));
-  if (database)
+  if (database && G_IS_TLS_DATABASE_GNUTLS (database))
     {
       priv->creds = g_tls_database_gnutls_get_credentials (G_TLS_DATABASE_GNUTLS (database), &my_error);
       if (!priv->creds)
@@ -993,6 +999,7 @@ g_tls_connection_gnutls_verify_chain (GTlsConnectionBase       *tls,
   GTlsCertificateFlags errors = 0;
   const char *hostname = NULL;
   char *free_hostname = NULL;
+  GTlsDatabase *database;
   guint gnutls_result;
   int ret;
 
@@ -1005,12 +1012,27 @@ g_tls_connection_gnutls_verify_chain (GTlsConnectionBase       *tls,
    * (a) is done by g_tls_database_verify_chain() and implemented using one of
    * several different functions of gnutls_x509_trust_list_t, e.g.
    * gnutls_x509_trust_list_verify_crt2() or one of the related functions.
-   *
-   * (b) is what we're doing here. The recommended way is to use
-   * gnutls_session_set_verify_cert(), but we can't do that because that would
-   * leave no way to implement the accept-certificate signal. The other way is
-   * to use gnutls_certificate_verify_peers3() or one of the related functions.
-   * This adds additional smarts that are not possible when using GTlsDatabase
+   * This is the best we can do if we have to use a GTlsDatabase that is not a
+   * GTlsDatabaseGnutls.
+   */
+  database = g_tls_connection_get_database (G_TLS_CONNECTION (gnutls));
+  if (!G_IS_TLS_DATABASE_GNUTLS (database))
+    {
+      return g_tls_database_verify_chain (database,
+                                          chain,
+                                          G_IS_TLS_CLIENT_CONNECTION (tls) ? G_TLS_DATABASE_PURPOSE_AUTHENTICATE_SERVER : G_TLS_DATABASE_PURPOSE_AUTHENTICATE_CLIENT,
+                                          identity,
+                                          g_tls_connection_get_interaction (G_TLS_CONNECTION (tls)),
+                                          G_TLS_DATABASE_VERIFY_NONE,
+                                          NULL,
+                                          error);
+    }
+
+  /* Now for (b). The recommended way is gnutls_session_set_verify_cert(), but
+   * we can't use that because that would leave no way to implement the
+   * GTlsConnection::accept-certificate signal. The other way is to use
+   * gnutls_certificate_verify_peers3() or one of the related functions. This
+   * adds additional smarts that are not possible when using GTlsDatabase
    * directly. For example, it checks name constraints, key usage, and basic
    * constraints. It also checks for stapled OCSP responses. Verification will
    * fail if the OCSP response indicates the certificate has been revoked.
@@ -1077,6 +1099,8 @@ static gchar *
 get_ciphersuite_name (gnutls_session_t session)
 {
   gnutls_protocol_t protocol_version = gnutls_protocol_get_version (session);
+  char *cipher_name;
+  char *result;
 
   if (protocol_version <= GNUTLS_TLS1_2 ||
       (protocol_version >= GNUTLS_DTLS0_9 && protocol_version <= GNUTLS_DTLS1_2))
@@ -1086,9 +1110,19 @@ get_ciphersuite_name (gnutls_session_t session)
                                                      gnutls_mac_get (session)));
     }
 
-  return g_strdup_printf ("TLS_%s_%s",
-                          gnutls_cipher_get_name (gnutls_cipher_get (session)),
-                          gnutls_digest_get_name (gnutls_prf_hash_get (session)));
+  cipher_name = g_strdup (gnutls_cipher_get_name (gnutls_cipher_get (session)));
+  for (char *c = cipher_name; *c != '\0'; c++)
+    {
+      if (*c == '-')
+        *c = '_';
+    }
+
+  result = g_strdup_printf ("TLS_%s_%s",
+                            cipher_name,
+                            gnutls_digest_get_name (gnutls_prf_hash_get (session)));
+  g_free (cipher_name);
+
+  return result;
 }
 
 static void
@@ -1138,6 +1172,16 @@ gnutls_get_binding (GTlsConnectionGnutls      *gnutls,
 
   if (ret == GNUTLS_E_SUCCESS)
     {
+      /* Older GnuTLS versions are known to return SUCCESS and empty data for TLSv1.3 tls-unique binding.
+       * While it may look prudent to catch here that specific corner case, the empty binding data is
+       * definitely not a SUCCESS, regardless of the version and type. */
+      if (cb.size == 0)
+        {
+          g_set_error (error, G_TLS_CHANNEL_BINDING_ERROR, G_TLS_CHANNEL_BINDING_ERROR_GENERAL_ERROR,
+                       _("Empty channel binding data indicates a bug in the TLS library implementation"));
+          return FALSE;
+        }
+
       if (data != NULL)
         {
           g_tls_log_debug (gnutls, "binding size %d", cb.size);
